"""
Data wrangling functions, typically with pandas
"""

import random
from typing import Literal, Union, Callable

import numpy as np
import pandas as pd
import operator as pyop


def send_column_to(
    dat: pd.DataFrame,
    move_cols: Union[list[str], str],
    send_to: Literal["front", "back"] = "front",
) -> pd.DataFrame:
    """
    Sends a column / columns to the front / back of a dataframe
    """
    if isinstance(move_cols, str):
        move_cols = [move_cols]

    leftovers = [col for col in dat.columns if col not in move_cols]
    if send_to == "front":
        cols = move_cols + leftovers
    elif send_to == "back":
        cols = leftovers + move_cols
    else:
        raise Exception(f"send_to must be 'front' or 'back', not {send_to}")
    return dat[cols]


def compute_distinct_values(dat: pd.DataFrame) -> pd.DataFrame:
    """Gets value counts of all non-numeric columns in a dataframe."""
    return (
        dat.select_dtypes(exclude=[np.number])
        .melt()
        .groupby("variable")["value"]
        .value_counts()
        .reset_index()
        .sort_values(by=["variable", "count"], ascending=False)
        .set_index("variable")
    )


def compute_pct_unique(series: pd.Series) -> float:
    """Returns the % of unique values of a series. This function will attempt
    to convert the series argument to a series.
    """
    if not isinstance(series, pd.Series):
        print(f"Attempting to convert {type(series)} to a {pd.Series}")
        series = pd.Series(series)
    return series.nunique() / series.size


def crosstab(dat: pd.DataFrame, group_by_vars, count_var) -> pd.DataFrame:
    """Computes pct / counts of count_var by group_by_vars

    crosstab(grouped, ['sex'], 'category')
    """
    a = dat.groupby(group_by_vars)[count_var].value_counts(normalize=True).reset_index()
    b = dat.groupby(group_by_vars)[count_var].value_counts().reset_index()
    return a.merge(b, on=group_by_vars + [count_var])


def sort_col_manually(input_df: pd.DataFrame, col_name: str, ordered_values: list):
    """Sorts a column in a dataframe in a manual order

    Example
    -------
    sort_col_manually(my_data, 'variable', ['first', 'second', 'third'])
    """
    all_values = input_df[col_name].unique()
    # Enable partial ordering (just moving things to the front)
    ordered_values = ordered_values + [
        val for val in all_values if val not in ordered_values
    ]
    input_df[col_name] = input_df[col_name].astype(
        pd.api.types.CategoricalDtype(categories=ordered_values, ordered=True)
    )
    input_df.sort_values(by=col_name, inplace=True)
    return input_df


def _try_get_operator(operator_name: str) -> Callable:
    try:
        return getattr(pyop, operator_name)
    except AttributeError:
        raise Exception(f'Could not find operator with name: "{operator_name}"')


def cols_with_n_distinct_values(df, n_unique, op_name: str = "le") -> pd.Series:
    """Shows columns with a certain number of unique values"""
    operation_fn = _try_get_operator(op_name)
    counts = df.apply(pd.Series.nunique)
    return counts[operation_fn(counts, n_unique)]


def filter_cols(
    data,
    n_distinct_thresh=3,
    selected_dtypes=None,
    applicator="both",
) -> pd.DataFrame:
    """Filters a dataframe based on distinct counts, OR using select_include"""
    dis_cols = set(cols_with_n_distinct_values(data, n_distinct_thresh).index.to_list())

    obj_cols = {}
    if selected_dtypes:
        obj_cols = set(data.select_dtypes(include=selected_dtypes).columns)

    cols_to_include = list(dis_cols.union(obj_cols))

    if not cols_to_include:
        raise Exception("That filtering didnt return anything!")

    return data[cols_to_include]


def shout(df: pd.DataFrame, msg: str = None) -> pd.DataFrame:
    """A simple function to be used with ``pd.pipe`` to print out
    the size of a dataframe and an optional message.

    :param df: The input dataframe
    :param msg: The message you want to print

    :returns: The original dataframe

    Example:
        .. ipython:: python

            from ourtils.wrangling import shout
            output = (
                pd.DataFrame([{'a': 10}, {'a': 15}, {'a': 20}])
                .pipe(shout, 'Starting pipeline')
                .loc[lambda x: x['a'] >= 15]
                .pipe(shout, 'After filtering')
            ); output
    """
    msg = f"{df.shape}: {msg or ''}"
    print(msg)
    return df


def collapse_multiindex(df: pd.DataFrame, sep: str = "_") -> pd.DataFrame:
    """Collapses a multi-index, this usually happens after some sort of aggregation.

    Currently only supports an index that's nested 1 level (so 2 levels)

    :param df: The input dataframe
    :param sep: A delimiter to use when joining the index values
    """
    _df = df.copy()
    index = _df.columns
    nlevels = index.nlevels
    assert nlevels == 2, f"Collapsing {nlevels} levels isn't handled yet."
    assert type(index) is pd.MultiIndex, "You must pass a dataframe with a multi-index."
    _df.columns = [sep.join([str(x) for x in v]) for v in index.values]
    _df.reset_index(inplace=True)
    return _df


def squish(
    df: pd.DataFrame,
    index_var: Union[str, list[str]],
    col_sep: str = "_",
    agg_func: Callable = list,
) -> pd.DataFrame:
    """Reshapes wide data into long format and adds a "group" column.

    :param df: The input dataframe
    :param index_var: The column or columns that uniquely identify
    :param col_sep: The thing to split the columns on
    :param agg_func: The function to use to aggregate the values. Defaults to a simple list

    Example:
        .. ipython:: python

            import pandas as pd
            from ourtils.wrangling import squish
            df = pd.DataFrame(
                columns=['index_var', 'a_1', 'a_2', 'b_1', 'b_2', 'b_3'],
                data=[
                    (1, 2, 3, 4, 5, 6),
                    (10, 20, 30, 40, 50, 60)
                ]
            )
            df

            df.pipe(squish, 'index_var')
    """
    if not isinstance(index_var, list):
        index_var = [index_var]

    def _try_split_column(colname: str) -> Union[str, None]:
        try:
            return colname.split(col_sep)[-2]
        except IndexError:
            return colname

    return (
        df.melt(id_vars=index_var, value_name="value", var_name="variable")
        .assign(group=lambda x: x["variable"].apply(_try_split_column))
        .groupby(index_var + ["group"], group_keys=False)["value"]
        .apply(lambda x: agg_func(x))
        .reset_index()
    )


def filter_random(df: pd.DataFrame, col: str) -> pd.DataFrame:
    """Returns the dataframe filtered to a random value of col.

    :param df: The input dataframe
    :param col: The column to pick a random value from
    """
    val = random.choice(df[col])
    return df.loc[lambda x: x[col] == val]


def create_column(
    df: pd.DataFrame, colname: str, func: Callable, *args, **kwargs
) -> pd.DataFrame:
    """Creates a new column using a function that takes column names as strings.

    :param df: The input dataframe
    :param colname: The name of the column you want to create
    :param func: The function to apply to the columns
    :param args: Column names to pass into ``func``


    Example:
        .. ipython:: python

            from ourtils.wrangling import create_column
            df = pd.DataFrame({
                'first': ['myfirst'],
                'last': ['mylast']
            })
            def create_name(first: str, last: str) -> str:
                return f'{last}, {first}'

            df.pipe(create_column, 'mynewcolumn', create_name, 'first', 'last')
    """
    return df.assign(
        # Need to do x.apply(lambda x: func(x['col1'], x['col2'], ..., mykwarg=15))
        __newcol=lambda x: x.apply(
            lambda x: func(
                *[x[arg] for arg in args],
                **{kwarg: value for kwarg, value in kwargs.items()},
            ),
            axis=1,
        )
    ).rename(columns={"__newcol": colname})


class ColumnSpec:
    """
    Specification for an individual column.
    """

    def __init__(
        self,
        col_names: Union[str, list[str]],
        mapping: dict,
        use_numeric_order=False,
        cat_overrides=None,
    ):
        if isinstance(col_names, str):
            self.col_names = [col_names]
        else:
            self.col_names = col_names
        self.mapping = mapping
        self.use_numeric_order = use_numeric_order
        self.cat_overrides = cat_overrides

    def as_category(self) -> pd.Categorical:
        if self.cat_overrides is not None:
            categories = self.cat_overrides
        else:
            categories = self.mapping.values()
        return pd.CategoricalDtype(categories, ordered=self.use_numeric_order)

    def apply_to_series(self, series: pd.Series) -> pd.Series:
        cat_type = self.as_category()
        return series.map(self.mapping).astype(cat_type)


class SpecCollection:
    """A collection of columnspecs."""

    def __init__(self, specs: list[ColumnSpec]):
        self.specs = specs

    def __getitem__(self, key) -> ColumnSpec:
        try:
            spec = [spec for spec in self.specs if key in spec.col_names][0]
        except IndexError:
            raise KeyError(f"No spec found for '{key}'")
        return spec

    def _map_to_series(self, series: pd.Series) -> pd.Series:
        try:
            spec = self[series.name]
            return spec.apply_to_series(series)
        except KeyError:
            return series

    def map_to_dataframe(self, data: pd.DataFrame) -> pd.DataFrame:
        """Applies all column specs to a dataframe."""
        return data.apply(self._map_to_series)


if __name__ == "__main__":
    dat = pd.DataFrame(
        {"grp": ["first", "second", "second", "third"], "score": [1, 1, 2, 2]}
    )
    sort_col_manually(dat, "grp", ["second", "third"])
    print(dat)



"""
Other misc items
"""

import os
from pathlib import Path

import pandas as pd


def create_analysis_notes(filename: Path, df: pd.DataFrame) -> None:
    """Creates a template markdown file to start making notes"""
    notes = [
        "# Analysis notes",
        "",
        "## Key Analysis Questions",
        "",
        "## Key Findings",
        "",
        "## Outstanding Questions",
        "",
        "## Dataset Info",
        "",
        f"Shape: {df.shape}",
        "",
        df.describe().to_markdown(),
        "",
        "## Columns",
        "",
    ]
    for col in df.columns:
        notes.append(f"- {col} ({df[col].dtype}): ")

    notes += [
        "",
        "## Other Notes",
        "",
    ]
    if os.path.exists(filename):
        raise Exception(f"'{filename}' already exists!!")
    with open(filename, "w") as f:
        f.writelines("\n".join(notes))



from collections import defaultdict
from dataclasses import dataclass
from typing import Callable
import pandas as pd


@dataclass
class SetComparison:
    only_a: set
    only_b: set
    in_both: set


def compare_sets(a: set, b: set, report=True) -> SetComparison:
    """Returns a tuple of useful differences of a set.

    Example:
        .. ipython:: python

            a = {1, 2, 3}
            b = {2, 3, 5}
            diffs.compare_sets(a, b)
    """
    if isinstance(a, list):
        a = set(a)
    if isinstance(b, list):
        b = set(b)

    in_both = a.intersection(b)
    only_a = a.difference(b)
    only_b = b.difference(a)
    if report:
        print(f"Only A: {only_a}")
        print(f"Only B: {only_b}")
        print(f"In both: {in_both}")

    return SetComparison(only_a, only_b, in_both)


class DataFrameDiffer:
    """
    A class to help compare dataframes. Deals with:

    - Matching columns
    - Filtering to differing rows

    Example:
        .. ipython:: python

            import pandas as pd
            from ourtils import diffs

            df1 = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
            df1

            df2 = pd.DataFrame({"a": [1, 2, 4], "B": [4, 5, 1], "c": [5, 6, 1]})
            df2

            # Create a DataFrameDiffer object
            diffy = diffs.DataFrameDiffer(df1, df2, "a")
            print(diffy.create_report())
            diffy.combined

    """

    lsuffix = "__left"
    rsuffix = "__right"

    @classmethod
    def create_diff(cls, row: pd.Series, cols: list[str]) -> dict:
        """Creates a dictionary that stores changes to the columns in a row.

        :param row: A row from a dataframe
        :param cols: A list of strings
        """
        _diff = defaultdict()
        for col in cols:
            b, a = row[f"{col}{cls.lsuffix}"], row[f"{col}{cls.rsuffix}"]
            if b != a:
                _diff[col] = (b, a)
        return _diff

    def __init__(
        self,
        left: pd.DataFrame,
        right: pd.DataFrame,
        join_on: list[str],
        column_cleaner: Callable = lambda x: x.strip().lower(),
    ):
        self.ldata = left.copy()
        self.rdata = right.copy()

        self.join_on = join_on
        self.column_cleaner = column_cleaner

        self.ldata.columns = [column_cleaner(col) for col in self.ldata]
        self.rdata.columns = [column_cleaner(col) for col in self.rdata]

        self.join_on = [column_cleaner(key) for key in self.join_on]

        self.columns_to_compare = [
            col for col in self.matching_columns if col not in self.join_on
        ]

        def create_message(_merge_indicator: str, diff: dict) -> str:
            """
            Creates a label based on the merge indicator as well as the changes
            that have happened to the row.
            """
            if _merge_indicator == "left_only":
                return "deleted"
            elif _merge_indicator == "right_only":
                return "inserted"
            elif diff:
                return "updated"
            return "same"

        self.combined = (
            pd.merge(
                self.ldata,
                self.rdata,
                on=self.join_on,
                indicator=True,
                how="outer",
                suffixes=(self.lsuffix, self.rsuffix),
            )
            .assign(
                diff=lambda x: x.apply(
                    lambda x: self.create_diff(x, self.columns_to_compare), axis=1
                )
            )
            .assign(
                action=lambda x: x.apply(
                    lambda x: create_message(x["_merge"], x["diff"]), axis=1
                )
            )
            .drop(columns=["_merge"])
        )

    @property
    def left_columns(self) -> set:
        return set(self.ldata.columns)

    @property
    def right_columns(self) -> set:
        return set(self.rdata.columns)

    @property
    def matching_columns(self) -> set:
        return self.left_columns.intersection(self.right_columns)

    @property
    def new_columns(self):
        return self.right_columns.difference(self.left_columns)

    @property
    def missing_columns(self):
        return self.left_columns.difference(self.right_columns)

    @property
    def comparable(self) -> pd.DataFrame:
        return self.combined

    @property
    def changed_data(self) -> pd.DataFrame:
        return self.combined.loc[lambda x: x["action"] != "same"]

    @property
    def summary_dict(self) -> dict:
        return defaultdict(
            int, self.combined["action"].value_counts(dropna=False).to_dict()
        )

    def _trim_message(self, message: str) -> str:
        return "\n".join([x.strip(" ") for x in message.splitlines() if x.strip(" ")])

    @property
    def n_inserted(self):
        return self.summary_dict["inserted"]

    @property
    def n_updated(self):
        return self.summary_dict["updated"]

    @property
    def n_deleted(self):
        return self.summary_dict["deleted"]

    @property
    def n_same(self):
        return self.summary_dict["same"]

    @property
    def summary_msg(self) -> str:
        return self._trim_message(
            f"""
        Inserted: {self.n_inserted}
        Updated: {self.n_updated}
        Deleted: {self.n_deleted}
        Same: {self.n_same}
        """
        )

    @property
    def n_changed_rows(self):
        return self.n_inserted + self.n_deleted + self.n_updated

    def create_report(self) -> str:
        """
        Returns a string report of the differences.
        """
        report = f"""
        {'-' * 15}
        ourtils.DifferenceReport
        {'-' * 15}
        Column summary:
        Removed: {self.missing_columns or ''}
        Added: {self.new_columns or ''}
        Matching: {self.matching_columns or ''}
        {'-' * 15}
        Row summary ({self.n_changed_rows} total changes):
        {self.summary_msg}
        """
        trimmed_report = self._trim_message(report)
        return trimmed_report



"""
All things modeling
"""

from functools import cached_property

import pandas as pd
import seaborn as sns
import patsy
from matplotlib import pyplot as plt

from statsmodels.regression import linear_model as lm
from statsmodels.formula import api as smf
from statsmodels.api import graphics as smg

from scipy import stats

from sklearn.pipeline import make_pipeline
from sklearn.model_selection import cross_validate


def standardize(numbers: pd.Series) -> pd.Series:
    """Safely standardizes a series."""
    if numbers.dtype == "object":
        return numbers
    return pd.Series(stats.zscore(numbers))


class RegressionResult:
    """Represents a single model"""

    def __init__(self, key: str, formula: str, description: str, data: pd.DataFrame):
        self.key = key
        self.description = description
        self.formula = formula
        self.data = data

    @property
    def y_variable(self) -> str:
        if len(self.model_descr.lhs_termlist) != 1:
            raise Exception("Multiple ys found in formula, not handling...")
        return self.model_descr.lhs_termlist[0].name()

    @cached_property
    def fitted_model(self) -> lm.RegressionResultsWrapper:
        return smf.ols(self.formula, data=self.data).fit()

    @cached_property
    def standardized_fitted_model(self):
        """Standardizes data before fitting."""
        standardized_data = self.data.apply(standardize)
        return smf.ols(self.formula, data=standardized_data).fit()

    @property
    def model_descr(self):
        return patsy.ModelDesc.from_formula(self.formula)

    @property
    def summary(self):
        return self.fitted_model.summary()

    @cached_property
    def regression_influences(self):
        return self.fitted_model.get_influence()

    def tag_dataframe(self, _df) -> pd.DataFrame:
        return _df.assign(
            key=self.key, description=self.description, formula=self.formula
        )

    def get_param_df(self, standardize=True) -> pd.DataFrame:
        if standardize:
            fitted_model = self.standardized_fitted_model
        else:
            fitted_model = self.fitted_model

        model_df = pd.DataFrame(
            {key: getattr(fitted_model, key) for key in ("params", "bse", "pvalues")}
        )
        model_df = self.tag_dataframe(model_df)
        model_df.reset_index(inplace=True)
        model_df.rename(columns={"index": "coef"}, inplace=True)
        return model_df

    @property
    def diagnostic_df(self) -> pd.DataFrame:
        _data = self.data.copy()

        # Distance from average, could be extreme X and Y...
        leverages = self.regression_influences.hat_matrix_diag

        # A measure of how much the model changes without this point
        cooks_values, cooks_ps = self.regression_influences.cooks_distance

        _data["predicted"] = self.fitted_model.fittedvalues
        _data["residual"] = self.fitted_model.resid
        _data["leverage"] = leverages
        _data["cooks_d"] = cooks_values
        _data["cooks_pvalue"] = cooks_ps
        _data = self.tag_dataframe(_data)
        return _data

    @property
    def info_df(self) -> pd.DataFrame:
        base_dict = {
            "key": self.key,
            "formula": self.formula,
            "description": self.description,
        }
        fitted_model_attrs = (
            "rsquared",
            "rsquared_adj",
            "nobs",
            "mse_model",
            "ssr",
            "mse_resid",
            "fvalue",
            "f_pvalue",
        )
        for attr in fitted_model_attrs:
            base_dict[attr] = getattr(self.fitted_model, attr)

        return pd.DataFrame.from_dict([base_dict])

    def get_n_most_influential_points(self, n=5):
        return self.diagnostic_df.sort_values(by="cooks_d", ascending=False).head(n)

    def plot_scatter(self, x_val: str, y_val=None, highlight_influence=False, **kwargs):
        if y_val is not None:
            y = y_val
        else:
            y = self.y_variable

        sns.lmplot(
            x=x_val,
            y=y,
            data=self.data,
            x_jitter=0.25,
            y_jitter=0.25,
            scatter_kws={"alpha": 0.3},
            **kwargs,
        )
        if highlight_influence:
            sns.scatterplot(
                data=self.data.loc[
                    lambda x: x.index.isin(self.idx_most_influential(5))
                ],
                x=x_val,
                y=y,
                color="red",
            )

    def plot_coefs(self, *args, **kwargs):
        bar_data = self.get_param_df(*args, **kwargs).query('coef != "Intercept"')
        axes = sns.barplot(data=bar_data, x="coef", y="params", hue="coef")
        for i in range(len(bar_data)):
            axes.errorbar(
                x=i,
                y=bar_data["params"].iloc[i],
                yerr=bar_data["bse"].iloc[i],
                fmt="none",
                color="black",
                capsize=5,
            )
        axes.axhline(y=0, color="black")

    def idx_most_influential(self, n):
        return self.get_n_most_influential_points(n).index

    def plot_n_influential(self, x_val: str, n: int, title=None, ax=None) -> None:
        """Plots the n most influential points."""
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = plt.gcf()

        if title is not None:
            ax.set_title(title)
        else:
            ax.set_title(f"{n} most influential points ('{self.y_variable}')")

        idx_most_influential = self.idx_most_influential(n)
        standard_points = self.data.loc[lambda x: ~x.index.isin(idx_most_influential)]
        influentials = self.data.loc[lambda x: x.index.isin(idx_most_influential)]

        min_x = self.data[x_val].min()
        max_x = self.data[x_val].max()
        ax.set_xlim([min_x, max_x])
        sns.stripplot(
            standard_points, x=x_val, y=self.y_variable, alpha=0.3, jitter=0.5, ax=ax
        )
        sns.stripplot(
            influentials,
            x=x_val,
            y=self.y_variable,
            color="red",
            ax=ax,
            alpha=1,
            jitter=1,
        )
        return fig, ax, idx_most_influential

    def plot_qq(self):
        smg.qqplot(self.fitted_model.resid)

    def plot_leverage_vs_resids(self):
        smg.plot_leverage_resid2(self.fitted_model)

    def plot_partial_regression(self):
        smg.plot_partregress_grid(self.fitted_model)


class RegResultCollection:
    """Represents a collection of regression models."""

    def __str__(self):
        return f"Regression results for {len(self.reg_results)} model(s)."

    @classmethod
    def build_from_mapper(cls, input_data: pd.DataFrame, model_mapper: list[tuple]):
        """
        Creates a RegResultCollection from a dictionary

        model_mapper: A dictionary of:
        ```python
        import pandas as pd
        {
            (key, formula): model description
        }
        ```
        input_data: The input dataframe to fit the model on
        """
        all_results = []
        seen_keys = set()
        for key, formula, description in model_mapper:
            if key in seen_keys:
                raise Exception(f"Duplicate model key '{key}' found!")
            all_results.append(RegressionResult(key, formula, description, input_data))
            seen_keys.add(key)

        return cls(all_results)

    def batch_plot_regs(self, x_var: str, n=10):
        fig, axes = plt.subplots(1, len(self.reg_results), figsize=(15, 5))
        for i, res in enumerate(self.reg_results):
            res.plot_n_influential(x_var, res.y_variable, n, True, ax=axes[i])
        return fig, axes

    def __init__(self, reg_results: list[RegressionResult]):
        self.reg_results = reg_results

    def __getitem__(self, item) -> RegressionResult:
        """Returns a reg result"""
        try:
            return self._models[item]
        except KeyError as e:
            raise Exception(f"No model with key={e}")

    def coef_dataframe(self, *args, **kwargs) -> pd.DataFrame:
        """Returns model summaries as a dataframe"""
        output = pd.concat(
            [result.get_param_df(*args, **kwargs) for result in self.reg_results]
        ).sort_values(by="coef")
        return output

    def _concat_frames(self, result_attr_name: str) -> pd.DataFrame:
        return pd.concat(
            [getattr(result, result_attr_name) for result in self.reg_results]
        )

    @property
    def model_summary(self):
        return self._concat_frames("info_df")

    @property
    def rowlevel_data(self) -> pd.DataFrame:
        return self._concat_frames("diagnostic_df")

    @property
    def _models(self) -> dict:
        return {result.key: result for result in self.reg_results}

    def plot_coefs(self, show_nums=True, scale_by=None, *args, **kwargs):
        """Plots the coefficients of each model as a bar chart with error bars."""
        bar_data = (
            self.coef_dataframe(*args, **kwargs).query('coef != "Intercept"')
            # ax.patches are in formula order
            .sort_values(by="formula")
        )

        if scale_by:
            bar_data["params"] = bar_data["params"] * scale_by

        n_total_bars = bar_data[["key", "coef"]].drop_duplicates().shape[0]

        axes = sns.barplot(data=bar_data, x="coef", y="params", hue="formula")
        axes.axhline(y=0, color="black")

        if scale_by:
            axes.set_title(f"Scaled by {scale_by}")

        # Remove any bonus patches that may get drawn
        correct_patches = [
            patch for patch in axes.patches if patch.get_x() and patch.get_height()
        ]
        if n_total_bars != len(correct_patches):
            raise Exception(
                f"Should see {n_total_bars} bar(s), but am seeing {len(correct_patches)} bar(s)"
            )

        if not show_nums:
            return axes

        def sign(x):
            if x < 0:
                return -1
            elif x > 0:
                return 1
            return 0

        for i, patch in enumerate(correct_patches):
            row_data = bar_data.iloc[i]
            bse = row_data["bse"]
            estimate = row_data["params"]
            x = patch.get_x()
            y = patch.get_height()
            width = patch.get_width()
            x_val = x + (width / 2)
            axes.text(
                x_val,
                sign(estimate) * 0.02 * (scale_by or 1),
                f"{estimate:.2f}",
                ha="center",
                va="center",
            )
            if scale_by:
                # Skip error bars if scaling
                continue
            axes.errorbar(x_val, y, yerr=bse, color="black", capsize=5)
            axes.text(
                x_val,
                y + (sign(estimate) * (bse + 0.02)),
                f"{bse:.3f}",
                ha="center",
                va="center",
            )

        return axes


class Sklearner:
    def __init__(self, X, y, preprocessing_pipeline, models: list):
        self.X = X
        self.y = y
        self.preprocessing_pipeline = preprocessing_pipeline
        self.models = models

    def create_cv_results(self, *args, **kwargs) -> pd.DataFrame:
        output = []
        for model in self.models:
            pipeline_with_model = make_pipeline(self.preprocessing_pipeline, model)
            results = cross_validate(
                pipeline_with_model, self.X, self.y, *args, **kwargs
            )
            _res = pd.DataFrame(results)
            _res["_model"] = str(model)
            output.append(_res)

        return pd.concat(output)



"""
Python only helpers
"""

from datetime import datetime


def pathsafenow() -> str:
    """Convert the current datetime into a safe string to be used as a directory name, useful
    when stashing data.

    Returns:
        str: A safe directory name string representing the current datetime.

    Example:
        .. ipython:: python

            from ourtils.general import pathsafenow
            pathsafenow()
    """
    now = datetime.now()
    formatted_datetime = now.strftime("%Y-%m-%d_%H-%M-%S")
    return formatted_datetime


def print_params(obj, *args) -> None:
    """Prints out the value of each arg from obj


    Example:
        .. ipython:: python

            from ourtils.general import print_params
            class Person:
                def __init__(self, name, age):
                    self.name = name
                    self.age = age
                def say_hello(self, formal=False):
                    if formal:
                        return f'Welcome, {self.name}.'
                    else:
                        return f'Hey {self.name}!'
            person = Person('spongebob', 100)
            print_params(person, 'name', 'age', 'say_hello()', {'say_hello': {'formal': True}})
    """
    print(f"Summary for {obj}")
    for arg in args:
        if isinstance(arg, dict):
            keys = list(arg.keys())
            if len(keys) > 1:
                raise ValueError("Method dicts are expected to have a single key!")
            method_name = keys[0].replace("()", "")
            method_kwargs = arg[method_name]
            val = getattr(obj, method_name)(**method_kwargs)
        elif arg.lower().endswith("()"):
            method_name = arg.replace("()", "")
            val = getattr(obj, method_name)()
        else:
            val = getattr(obj, arg)
        print(f"{arg}: {val}")


def qatl(obj) -> list:
    """Quick attribute list, filters out private / dunders"""
    return [attr for attr in dir(obj) if not attr.startswith("_")]



from datetime import datetime
import tracemalloc
from contextlib import contextmanager
import pandas as pd


@contextmanager
def ExplainBlock(msg, width=50, char="*"):
    """Wraps code in a title block"""
    msg = f" {msg} "
    leftover_chars = width - len(msg)
    left_chars = leftover_chars // 2
    if max(left_chars, 0):
        right_chars = leftover_chars - left_chars
    else:
        right_chars = 0
    title_msg = f"{left_chars*char}{msg}{right_chars*char}"
    print(title_msg)
    yield
    print(char * len(title_msg))


@contextmanager
def MemoryRecorder():
    """
    Prints out peak and current memory usage
    """
    start = datetime.now()
    tracemalloc.start()
    yield
    current_size, peak = tracemalloc.get_traced_memory()
    tracemalloc.stop()
    end = datetime.now()
    div_factor = 1024**2  # Convert to MiB
    current_size /= div_factor
    peak /= div_factor
    print(f"Peak={peak:.2f}MiB, Current={current_size:.2f}MiB, took {end-start}")


if __name__ == "__main__":
    my_list = []

    with MemoryRecorder() as recorder:
        df = pd.DataFrame({"col1": ["A", "B", "A", "C", "B", "C"] * 1000})
        for _ in range(5):
            my_list.append(df.copy())



"""
Anything to do with excel
"""


def excel_column_index(col_name: str) -> int:
    """
    Returns the column index (starting with 1) given an excel column name (e.g. "AA")
    - Thank you chatgpt
    """
    index = 0
    for char in col_name.upper():
        index = index * 26 + (ord(char) - ord("A") + 1)
    return index


def index_to_excel_column(index: int) -> str:
    """
    Returns the excel column name based on an index (should start at 1)
    - Thank you chatgpt
    """
    if index < 1:
        raise ValueError("Index must be a positive integer.")

    column_name = []
    while index > 0:
        index -= 1  # Adjust for 0-based indexing
        column_name.append(chr(index % 26 + ord("A")))
        index //= 26

    return "".join(reversed(column_name))



"""
All things data visualization
"""

from typing import Union
import plotly.express as px
import pandas as pd
from matplotlib import pyplot as plt
from .wrangling import compute_pct_unique, crosstab


def plot_pct_uniqueness(dat: pd.DataFrame) -> tuple:
    """
    Draws a plot that counts the unique values in a dataframe.
    Returns a tuple of the matplotlib figure / axis objects so they
    can be further customized
    """
    fig, ax = plt.subplots()
    unique_pct = dat.apply(compute_pct_unique)
    unique_pct.plot(kind="barh", xlim=(0, 1), ax=ax)
    # n unique values for each column
    pct_value_mapper = dat.apply(pd.Series.nunique).to_dict()
    ytick_labels = [tick.get_text() for tick in ax.get_yticklabels()]
    print(pct_value_mapper)
    for i, rect in enumerate(ax.patches):
        X_COORD = rect.get_width() + 0.05
        Y_COORD = rect.get_y() + rect.get_height() / 2
        pct_unique = pct_value_mapper[ytick_labels[i]]
        txt = f"{round(rect.get_width(), 3)} ({pct_unique})"
        ax.text(X_COORD, Y_COORD, txt, ha="left", va="center_baseline")

    ax.set_title(f"Uniqueness ({dat.shape[0]} total rows)")
    ax.set_xticks([0, 1])
    ax.tick_params(length=0)
    ax.set_xticklabels(["All Values Identical (0)", "All Values Unique (1)"])
    return fig, ax


def plot_pct(
    data,
    group_by_vars: Union[list, str],
    target_var: str,
    use_proportion=True,
    plotly_kwargs=None,
    auto_show=True,
) -> tuple:
    """Plots percentage counts by group. Returns"""
    plotly_kwargs = plotly_kwargs or {}
    facet_kwargs = {}

    if isinstance(group_by_vars, str):
        group_by_vars = [group_by_vars]

    y = group_by_vars[0]
    if len(group_by_vars) > 1:
        facet_kwargs["facet_col"] = group_by_vars[1]
    if len(group_by_vars) > 2:
        facet_kwargs["facet_row"] = group_by_vars[2]
    if len(group_by_vars) > 3:
        raise ValueError("Impossible to visualize > 3 groupings!")

    # Show all labels by default
    hd = {"proportion": ":.1%", "count": True}
    for col in group_by_vars:
        hd[col] = True

    summary_data = crosstab(data, group_by_vars, target_var)
    x_var = "proportion" if use_proportion else "count"
    fig = px.bar(
        summary_data, x=x_var, y=y, color=target_var, **facet_kwargs, **plotly_kwargs
    )
    if auto_show:
        fig.show()
    return fig, summary_data
